/*
 *	Copyright (C) 2007, 2008, 2009. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/*!
@file siftDetector.cpp
@ingroup OpenCVExamplePrograms
@brief Obtains <a href="http://en.wikipedia.org/wiki/Scale-invariant_feature_transform">SIFT descriptors</a> over the frames from an input video source.

This is an example of QVision application using OpenCV functionallity. The program obtains SIFT features on the frames from the input video sequence using <a href="http://web.engr.oregonstate.edu/~hess/index.html">Rob Hess's SIFT implementation</a>. The application displays the detected features on an image canvas:

@image html opencv-sift.png

@section Usage
Once compiled, the program should be used with this command line:
\code
	./siftextractor --URL=<path or url to video>
\endcode

@param URL Video source identificator. See dynamic property URL for @ref QVMPlayerCamera.
@author PARP Research Group. University of Murcia, Spain.
*/

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <QDebug>

#include <QVApplication>
#include <QVMPlayerCameraWorker>
#include <QVDefaultGUI>
#include <QVImageCanvas>

#include <qvipp.h>

#include <sift.h>
#include <imgfeatures.h>
#include <utils.h>
#include <highgui.h>
#include <unistd.h>
#include <cxcore.h>
#include <cv.h>

#include <QVector>
#include <QList>

#ifndef DOXYGEN_IGNORE_THIS

// QVSIFTDescriptor
// should have:
//	- parameters for the ellipse
//	- scale of a Lowe-style feature
//	- orientation of a Lowe-style feature
//	- DESCRIPTOR (array of doubles)
//	- Type (Oxford or Lowe)
/**
Structure to represent an affine invariant image feature.  The fields
x, y, a, b, c represent the affine region around the feature:

a(x-u)^2 + 2b(x-u)(y-v) + c(y-v)^2 = 1
*/
class QVSIFTDescriptor
	{
	public:
		double x, y, a, b, c, scale, orientation;
		QVector<double> descriptor;
		int type;

		QVSIFTDescriptor(struct feature f)
			{
			this->x = f.x;
			this->y = f.y;
			this->a = f.a;
			this->b = f.b;
			this->c = f.c;
			this->orientation = f.ori;
			this->scale = f.scl;
			Q_ASSERT_X(f.type == FEATURE_LOWE, "QVSIFTDescriptor", "Feature type not Lowe");
			this->type = f.type;
			for (int i = 0; i< FEATURE_MAX_D; i++)
				descriptor.append(f.descr[i]);
			}
	};

/*
@brief Find SIFT features in an image using user-specified parameter values.  All detected features are stored in the array pointed to by \a feat.

@param img the image in which to detect features
@param feat a pointer to an array in which to store detected features
@param intvls the number of intervals sampled per octave of scale space
@param sigma the amount of Gaussian smoothing applied to each image level before building the scale space representation for an octave
@param contr_thr a threshold on the value of the scale space function
	\f$\left|D(\hat{x})\right|\f$
where \f$\hat{x}\f$ is a vector specifying feature location and scale, used to reject unstable features;  assumes pixel values in the range [0, 1]
@param curv_thr threshold on a feature's ratio of principle curvatures used to reject features that are too edge-like
@param img_dbl should be 1 if image doubling prior to scale space construction is desired or 0 if not
@param descr_width the width, \f$n\f$, of the \f$n \times n\f$ array of orientation histograms used to compute a feature's descriptor
@param descr_hist_bins the number of orientations in each of the histograms in the array used to compute a feature's descriptor
@return Returns the number of keypoints stored in \a feat or -1 on failure
*/
const QList<QVSIFTDescriptor>	getSIFTDescriptors(const QVImage<uChar, 3> &image,
				int intvls = SIFT_INTVLS, double sigma = SIFT_SIGMA, double contr_thr = SIFT_CONTR_THR,
				int curv_thr = SIFT_CURV_THR, int img_dbl = SIFT_IMG_DBL, int descr_width = SIFT_DESCR_WIDTH,
				int descr_hist_bins = SIFT_DESCR_HIST_BINS)
	{
	IplImage *imageIPL = image;
	
	struct feature * feat;
	const int n = _sift_features(imageIPL, &feat, intvls, sigma, contr_thr, curv_thr, img_dbl, descr_width, descr_hist_bins);

	QList<QVSIFTDescriptor> result;
	for (int i = 0; i < n; i++)
		result.append(QVSIFTDescriptor(feat[i]));

	cvReleaseImage(&imageIPL);

	free(feat);
	return result;
	}

class SIFTDetectorWorker: public QVWorker
	{
		
	public:
		SIFTDetectorWorker(QString name): QVWorker(name)
			{
			addProperty< QVImage<uChar,3> >("Input image", inputFlag|outputFlag);
			addProperty< QList<QVPolyline> >("SIFT ellipses", outputFlag);
			}

		void iterate()
			{
			// 0. Get input image
			const QVImage<uChar,3> image = getPropertyValue< QVImage<uChar,3> >("Input image");
			timeFlag("Image grab");

			// 1. Get SIFT descriptors
			QList<QVSIFTDescriptor>	descriptors = getSIFTDescriptors(image);
			timeFlag("Get SIFT descriptors");

			// 2. Get SIFT features as ellipses
			QList<QVPolyline> SIFTEllipses;
			foreach (QVSIFTDescriptor descriptor, descriptors)
				SIFTEllipses.append(	QVPolyline::ellipse(
									20, descriptor.x, descriptor.y,
									sqrt(descriptor.scale)*5,
									sqrt(descriptor.scale)*2.5,
									-descriptor.orientation)
									);

			timeFlag("Draw SIFT descriptors");

			// 3. Publish ellipses
			setPropertyValue< QList<QVPolyline> >("SIFT ellipses", SIFTEllipses);
			timeFlag("Publish SIFT ellipses");
			}
	};

int main(int argc, char *argv[])
	{
	QVApplication app(argc, argv,
		"Example program for QVision library. Gets SIFT descriptors from image."
		);
	
	QVMPlayerCameraWorker camera("Video");
	SIFTDetectorWorker worker("SIFT Worker");
	camera.linkProperty(&worker,"Input image");

	QVDefaultGUI interface;

	QVImageCanvas imageCanvas("Sift image");
	worker.linkProperty("Input image", imageCanvas);
	worker.linkProperty("SIFT ellipses", imageCanvas);

	return app.exec();
	}
#endif
