/*
 *	Copyright (C) 2007, 2008, 2009. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvgenericimage.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVGENERICIMAGE_H
#define QVGENERICIMAGE_H

#include <qvdefines.h>
#include <QRect>
#include <QPoint>

/*! @class QVGenericImage qvgenericimage.h qvcore/qvgenericimage.h
@brief Base class for @ref QVImage objects.

For instance, this class is useful to create code like this:

\code
int imageAreaSize(QVGenericImage *image)
	{
	return image->getCols() * image->getRows();
	}
\endcode

Where it doesn't matter the bit depth or number of channels of the image, but the number of cols and rows of it, as well as generic containers, that can hold a variety of image types, for example:

\code
class GenericImageContainer
	{
	public:
		int count;
		QVGenericImage *images;
		[...]
		void addImage(QVGenericImage *image)	{ images[count++] = &image; }
		QVGenericImage getImage(int index)	{ return images[index]; }
	}
\endcode

Thus, this class offers access methods for all kind of information about the image, like dimensions (see @ref getCols() and @ref getRows() methods), bit depth and number of planes (see @ref getTypeSize() and @ref getChannels() methods), ROI and anchor of the image (see @ref getROI() and @ref getAnchor() methods mainly), but pixel accessor methods, that are defined in the subclass @ref QVImage, and depends on the bit depth and number of planes of the image.

@todo eliminar el sufijo 'get' de todas las funciones. getCols -> cols, getRows -> rows
@see QVImage
@ingroup qvcore
*/
class QVGenericImage
	{
	public:
		/// @brief Default constructor.
		///
		/// Initializes image ROI and anchor to zero offset and width, both.
		QVGenericImage (): roi(0,0,0,0), anchor(QPoint(0,0))					{ }

		/// @brief Copy constructor.
		///
		/// Initializes image ROI and anchor to ROI and anchor from image 'img'.
		QVGenericImage(QVGenericImage const &img): roi(img.getROI()), anchor(img.getAnchor())	{ }

		/// @brief Virtual destructor for QVGenericImage.
		virtual ~QVGenericImage ()								{ };

		/// @brief Virtual method for getting number of columns from image.
		///
		/// @returns number of columns for image,
		virtual uInt	getCols()		const	= 0;

		/// @brief Virtual method for getting number of rows from image.
		///
		/// @returns number of rows for image,
		virtual uInt	getRows()		const	= 0;

		/// @brief Virtual method for getting step from image.
		///
		/// This functions returns the distance in bytes between starts of consecutive lines in the image.
		///
		/// This may be not equal to a multiple of the number of columns, so this value is necessary to normal access
		/// to the image.
		///
		/// @returns distance in bytes between two consecutive lines in the image.
		virtual uInt	getStep()		const	= 0;

		/// @brief Virtual method for getting number of planes for the image.
		///
		/// Returned value should be equal to template parameter 'Channels', for any QVImage. This is the number
		/// of channels that can be accessed in the image.
		///
		/// @returns number of channels of the image.
		virtual uInt	getChannels()		const	= 0;

		/// @brief Virtual method for getting bit depth of the image, in bytes.
		///
		/// Returned value should be equal to size of template parameter 'Type', for any QVImage. This is the size
		/// of any element of the data array that stores the image in memory.
		///
		/// @returns number of channels of the image.
		virtual	uInt	getTypeSize()		const	= 0;

		/// @brief Virtual method for getting data size for the image.
		///
		/// This functions, for any image, returns the following value:
		/// \verbatim dataSize = rows * step * planes \endverbatim
		///
		/// @returns size in bytes of data buffer for the image.
		virtual	uInt	getDataSize()		const	= 0;

		/// @brief Method for obtaining actual region of interest for image.
		///
		/// The region of interest of an image is used for various operations over it. Check the manual for further
		/// documentation
		///
		/// @returns a QRect that contains the region of interest of the image.
		const QRect &	getROI()		const	{ return roi; }

		/// @brief Obtains actual anchor for image.
		///
		/// The anchor of an image is used for various operations over it. Check the manual for further
		/// documentation
		///
		/// @returns a QPoint that indicates anchor of the image.
		const QPoint &	getAnchor()		const	{ return anchor; }

		// ROI operations
		/// @brief Resets the region of interest of the image.
		///
		/// Makes the region of interest of the image hold the whole image.
		///
		/// That is, makes the region of interest be the rectangle which has upper left corner in pixel (0,0),
		/// and lower right corner in pixel (columns, rows), where columns and rows are the number of columns
		/// and rows of the image.
		///
		void resetROI()				{ setROI(0,0,getCols(), getRows()); }

		/// @brief Reduces the region of interest of the image, in a given amount of pixels.
		///
		/// Adds cols and rows parameters to the top left corner of the region of interest, and those values
		/// doubled, to the width.
		///
		/// That is, reduces the region of interest in a border, of columns and rows indicated in the parameters.
		///
		/// @param cols	columns which should be reduced from the border of the roi.
		/// @param rows	rows which should be reduced from the border of the roi.
		void erodeROI(uInt cols, uInt rows)
			{ setROI(getROI().x()+cols, getROI().y()+rows, getROI().width()-2*cols, getROI().height()-2*rows); }

		/// @brief Increments the region of interest of the image, in a given amount of pixels.
		///
		/// Substracts cols and rows parameters to the top left corner of the region of interest, and those values
		/// doubled, to the width.
		///
		/// That is, increments the region of interest in a border, of columns and rows indicated in the parameters.
		///
		/// @param cols	columns which should be incremented to the border of the roi.
		/// @param rows	rows which should be incremented to the border of the roi.
		void dilateROI(uInt cols, uInt rows)
			{ setROI(getROI().x()-cols, getROI().y()-rows, getROI().width()+2*cols, getROI().height()+2*rows); }

		/// @brief Sets dimensions and location of region of interest in an image.
		///
		/// Region of interest for an image is a rectangle significant to many operations that can be done over an
		/// image.
		///
		/// It is defined by pixel where top left corner for this rectangle is located, and width & height of
		/// this rectangle.
		///
		/// @param x column for the pixel where top left corner of the region of interest is located.
		/// @param y row for the pixel where top left corner of the region of interest is located.
		/// @param w width for the region of interest.
		/// @param h height for the region of interest.
		void setROI(int x, int y, uInt w, uInt h) { setROI(QRect(x,y,w,h)); }

		/// @brief Sets the region of interest for an image, holding the whole image, except a border of a given width.
		///
		/// @param margin size of the border.
		void setMarginROI(int margin)
			{ setROI(margin, margin, getCols() - 2*margin, getRows() -2*margin); }

		/// @brief Directly sets the region of interest for an image, with a given rectangle.
		///
		/// @param rect rectangle that represents region of interest.
		void setROI(const QRect &rect)
			{
			Q_ASSERT_X(rect.x() >= 0,"QVGenericImage::setROI()","QRect.x() is less than zero");
			Q_ASSERT_X(rect.y() >= 0,"QVGenericImage::setROI()","QRect.y() is less than zero");
			Q_ASSERT_X(rect.width() > 0,"QVGenericImage::setROI()","QRect.width() is less or equal to zero");
			Q_ASSERT_X(rect.height() > 0,"QVGenericImage::setROI()","QRect.height() is less or equal to zero");
			Q_ASSERT_X(rect.x()+rect.width() <= (int) getCols(),"QVGenericImage::setROI()","x + width > columns");
			Q_ASSERT_X(rect.y()+rect.height() <= (int) getRows(),"QVGenericImage::setROI()","y + height > rows");
			roi = rect;
			}

		// Anchor operations
		/// @brief Sets anchor to pixel (0,0).
		///
		void resetAnchor()			{ setAnchor(0,0); }

		/// @brief Sets anchor to a given pixel coordinates.
		///
		/// @param col column of the pixel where the anchor is located.
		/// @param row row of the pixel where the anchor is located.
		void setAnchor(int col, int row)	{ setAnchor(QPoint(col,row)); }

		/// @brief Directly sets anchor to a given pixel coordinates, given a point.
		///
		/// @param point QPoint where the anchor is located.
		void setAnchor(const QPoint &point)
			{
			Q_ASSERT_X(point.x()>=0,"QVGenericImage::setAnchor()","horizontal value for anchor is less than zero");
			Q_ASSERT_X(point.y()>=0,"QVGenericImage::setAnchor()","vertical value for anchor is less than zero");
			Q_ASSERT_X(point.x() < (int) getCols(),"QVGenericImage::setAnchor()","horizontal value exceeds cols");
			Q_ASSERT_X(point.y() < (int) getRows(),"QVGenericImage::setAnchor()","vertical value exceeds rows");
			anchor = point;
			}

		/// @brief Function to make introspection in the image, and obtain a char string identifying the type of the
		/// real image contained in the QVGenericImage.
		///
		/// @returns pointer to a constant char string identifying image type.
		virtual const char * getTypeQString() const = 0;

		/// @brief Function to check whether an image is compatible with given image type id char string.
		///
		/// This functions checks if this image has same bit depth and channel number, as the type of QVGenericImage
		/// represented by the parameter string.
		///
		/// Id char string can be that returned by getTypeQString() function, from another QVGenericImage, or
		/// one specified by the programmer.
		///
		/// This function is used to dynamic type introspection, to determine subclass of a QVGenericImage, and
		/// perform different functionality with it, depending on bit depth or number of planes. An example code for
		/// this use is below:
		/// \code [...]
		/// QVImage image;
		///
		/// [...]
		///
		/// if(image->isCompatibleWith("QVImage<uChar,1>"))
		///	// do stuff with the uChar bit depth, one channel image ...
		/// else if(image->isCompatibleWith("QVImage<uChar,3>"))
		///	// do stuff with the uChar bit depth, three channel image ...
		/// [...]
		/// else
		///	// alternate code for that type of image ...
		///
		/// [...] \endcode
		///
		/// @param qvimageClassName char string identifying QVGenericImage type of an image.
		/// @returns true if this image is compatible in type with that represented by the string.
		bool isCompatibleWith(const char *qvImageClassName) const
			{
			Q_ASSERT_X(qvImageClassName!= NULL,"QVGenericImage::isCompatibleWith()","class name string is NULL");
			return (0 == strcmp(this->getTypeQString(),qvImageClassName));
			}

		/// @brief Function to check whether an image is compatible with this image.
		///
		/// This functions checks if this image is compatible with a given image (that is, it has same bit depth and
		/// same number of channels).
		///
		/// Id char string can be that returned by getTypeQString() function, from another QVGenericImage, or
		/// one specified by the programmer.
		///
		/// @param image image to check type with.
		/// @returns true if this image is compatible in type with that represented by the string.
		bool isCompatibleWith(const QVGenericImage *image) const
			{
			Q_ASSERT_X(image!= NULL,"QVGenericImage::isCompatibleWith()","NULL pointer");
			return this->isCompatibleWith(image->getTypeQString());
			}

	protected:
		QRect roi;
		QPoint anchor;
	};


#endif // QVGENERICIMAGE_H
