/*
 *	Copyright (C) 2007, 2008, 2009. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvimagecanvas.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVIMAGECANVAS_H
#define QVIMAGECANVAS_H

#include <QPointF>
#include <QVWorker>
#include <QVPolyline>
#include <QVPolylineF>
#include <qvgui/qvcanvas.h>
#include <qvgui/qvplot.h>

/*!
@class QVImageCanvas qvimagecanvas.h qvgui/qvimagecanvas.h
@brief Displayer widget for @ref QVImage objects and other data types.

Displays images from an image source:

@image html qvimagecanvas_penguin.png

For further information about its usage, see section @ref QVImageCanvasWidget.
@ingroup qvgui
*/
class QVImageCanvas: public QVCanvas, public QVPropertyContainer
	{
		Q_OBJECT
	public:
		/// @brief Image canvas constructor
		///
		/// @param name Name of the image canvas
		/// @param parent parent QObject of the canvas
		QVImageCanvas(const QString name = QString(), QWidget *parent=0);

		/// @brief Sets the display color for a registered input property of the canvas
		///
		/// Registered properties of types QList<QPoint>, QList<QPointF>, QList<QVPolyline> and QList<QVPolylineF> are displayed in the canvas
		/// with a default color assigned by the link method. The user can specify a non-default color value using method @ref setColor.
		///
		/// The following code registers an image property, and a QList<QPointF> property in an image canvas from a source worker, and sets the
		/// display color of the points received from the worker, through the <i>Detected points</i> property to be red.
		///
		/// @code
		/// QVImageCanvas imageCanvas("Harris corners");
		/// harrisCornersDetector.linkProperty("Input image", imageCanvas);
 		/// // Property "Detected points" is of type QList<QpointF>
		/// harrisCornersDetector.linkProperty("Detected points", imageCanvas);
		/// imageCanvas.setColor("Detected points", Qt::red);
		/// @endcode
		///
		/// @param name Name of the source property to set the color.
		/// @param QColor Color to display the elements contained in the property.
		bool setColor(const QString &name, const QColor &color)
			{ return setPropertyValue<QColor>("Color for " + name, color); }

		bool setPrintTags(const QString &name, const bool &printTags)
			{ return setPropertyValue<bool>("Print tags for " + name, printTags); }

		/// @brief Sets the display radius for a registered input property of the canvas
		///
		/// Some registered properties, for example points, are displayed in the canvas with a default radius size of 3. 
		/// As the @ref setColor, this method can be used to specify a display radius size different than the default.
		///
		/// @param name Name of the source property to set the radius.
		/// @param radius Radius size of the property.
		bool setRadius(const QString &name, const int &radius)
			{ return setPropertyValue<int>("Radius for " + name, radius); }

		/// @brief Sets the saturation range to display floating point images
		///
		/// The default saturation range to display a floating point image (QVImage<sFloat, 1> or QVImage<sFloat, 3>) is [0, 255]. This means that
		/// pixel intensity values outside that range will be set to 0 or 255, wether the intensity is below or above that range.
		/// 
		/// This method can be used to specify a different saturation range.
		///
		/// @param low Lower value for the saturation range.
		/// @param high Higher value for the saturation range.
		void setSaturationRange(const float low, const float high);

		/// @brief Display user defined data types method
		///
		/// By reimplementing this method on @ref QVImageCanvas subclasses, the developer can modify the way the objects are displayed in the canvas, or
		/// specify the way new user data type objects must be displayed.
		/// 
		/// These subclasses can create image canvas objects which display registered objects the new way.
		/// 
		/// Method @ref custom_viewer is called whenever the canvas must be repainted, for example due to changes in the input properties or a
		/// canvas resizement. The reimplemented code for this method must read the user defined properties, and use @ref QVImageCanvas or QWidget
		/// functionallity to correctly display them.
		///
		/// An example @ref custom_viewer function reimplementation follows:
		///
		/// @code
		/// void MyImageCanvas::custom_viewer()
		/// 	{
		///	foreach(QString propertyName, getPropertyListByType< QList< QVPolylineF > >())
		///		{
		///		QList< QVPolylineF > polylineList = getPropertyValue< QList< QVPolylineF > >(propertyName);
		///		foreach(QVPolylineF polyline, polylineList)
		///			{
		///			// Set the color of the polyline
		///			const int pointHash = qHash(polyline.first());
		///			draw(polyline, qvColors[pointHash % 10]);
		///
		///			// Draw red dots on the end of the polyline
		///			getQVPainter()->setPen(Qt::red);
		///			getQVPainter()->setBrush(QBrush(Qt::red, Qt::SolidPattern));
		///			getQVPainter()->drawEllipse(polyline.first().x()-1, polyline.first().y()-1, 3, 3);
		///			getQVPainter()->drawEllipse(polyline.last().x()-1, polyline.last().y()-1, 3, 3);
		///			}
		///		}
		///	}
		/// @endcode
		///
		/// Image canvas objects created from the class <i>MyImageCanvas</i>, which supossely inherits from @ref QVImageCanvas will display polylines
		/// contained in input polyline lists with different colors, depending on the location of their first point, and red dots of radius 3 at the
		/// ends of each polyline.
		virtual void custom_viewer() { };

		void setLowHigh(float low,float high)
			{
			std::cout << "DEPRECATED, use setSaturationRange instead" << std::endl;
			setSaturationRange(low, high);
			}

		void unlink();

		#ifndef DOXYGEN_IGNORE_THIS
		void viewer();
		#endif

	public slots:
		void rectSelectedSlot(QRect rect);
		void polySelectedSlot(QPoint point, bool reset, TPolyMode mode);
		void circleSelectedSlot(QPoint center, float radius);

	protected:
		bool linkUnspecifiedInputProperty(QVPropertyContainer *sourceContainer, QString sourcePropName, LinkType linkType = AsynchronousLink);
		bool linkUnspecifiedOutputProperty(QVPropertyContainer *destContainer, QString destPropName, LinkType linkType = AsynchronousLink);
		bool treatUnlinkInputProperty(QString destPropName, QVPropertyContainer *sourceCont, QString sourcePropName);

		void draw(const QList<QPoint> &pointList, QColor color = Qt::red, bool printTags = false, int radius = 3);
		void draw(const QList<QPointF> &pointList, QColor color = Qt::red, bool printTags = false, double radius = 3);
		void draw(const QVPolyline &polyline, QColor color = Qt::red, bool printTags = false);
		void draw(const QVPolylineF &polylinef, QColor color = Qt::red, bool printTags = false);
		void draw(const QRect &rectangle, QColor color = Qt::red, bool printTags = false);

		void closeEvent(QCloseEvent *event) { Q_UNUSED(event); emit closed(); }
	private:

		const QColor	getNextColor()
			{
			QColor color = qvColors[colorCursor++];
			colorCursor %= 10;

			return color;
			}

		float _low,_high;
		int colorCursor, contentLinkedWorkers;

	signals:
		void closed();

	};

Q_DECLARE_METATYPE(TPolyMode);

#endif
