/*
 *	Copyright (C) 2007, 2008, 2009. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvmplayercamera.cpp
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>

#include <QApplication>
#include <QDebug>
#include <QObject>
#include <QStringList>
#include <QRegExp>
#include <QMutex>

#include <qvipp.h>
#include <qvio.h>

#include <QVWorker>
#include <QVMPlayerProxy>

/******************* Auxiliary QVCheckOKMPlayerCamera Class *******************/
QVCheckOKMplayerCamera::QVCheckOKMplayerCamera(QFile & fifo_file,int max_time_ms_to_wait_for_open) : QThread(), _fifo_file(fifo_file), _max_time_ms_to_wait_for_open(max_time_ms_to_wait_for_open)
	{
	qDebug() << "QVCheckOKMplayerCamera::QVCheckOKMplayerCamera()";
	moveToThread(this);
	start();
	qDebug() << "QVCheckOKMplayerCamera::QVCheckOKMplayerCamera() <- return";
	}

void QVCheckOKMplayerCamera::run()
	{
	qDebug() << "QVCheckOKMplayerCamera::run()";
	QTimer::singleShot(_max_time_ms_to_wait_for_open, this, SLOT(writeErrorInFifo()));
	qDebug() << "QVCheckOKMplayerCamera::run() <- entering event loop";
	exec();
	qDebug() << "QVCheckOKMplayerCamera::run() <- back from event loop";
	qDebug() << "QVCheckOKMplayerCamera::run() <- return";
	}


void QVCheckOKMplayerCamera::writeErrorInFifo()
	{
	qDebug() << "QVCheckOKMplayerCamera::writeErrorInFifo()";
	_fifo_file.write("MPLAYER ERROR\n"); 
	qDebug() << "QVCheckOKMplayerCamera::writeErrorInFifo() -> return";
	};

/********************* Auxiliary QVCameraThread Class *************************/
QVCameraThread::QVCameraThread(QObject *object,char *slot): QThread(), _object(object), _slot(slot)
	{
	qDebug() << "QVCameraThread::QVCameraThread()";
	_object->moveToThread(this);
	qDebug() << "QVCameraThread::QVCameraThread(): object moved to thread";
	qDebug() << "QVCameraThread::QVCameraThread() <- return";
	}

QVCameraThread::~QVCameraThread()
	{
	qDebug() << "QVCameraThread::~QVCameraThread()";
	disconnect();
	quit();
	qDebug() << "QVCameraThread::~QVCameraThread(): sent quit to thread";
	wait();
	qDebug() << "QVCameraThread::~QVCameraThread(): thread finished";
	qDebug() << "QVCameraThread::~QVCameraThread() <- return";
	}

void QVCameraThread::run()
	{
	qDebug() << "QVCameraThread::run()";
	QTimer *timer = new QTimer();
	timer->start(0);
	connect(timer, SIGNAL(timeout()),_object,_slot);
	qDebug() << "QVCameraThread::run(): timer created, started and connected";
	qDebug() << "QVCameraThread::run(): entering event processing loop";
	exec();
	qDebug() << "QVCameraThread::run(): back from event processing loop";
	delete timer;
	qDebug() << "QVCameraThread::run(): timer deleted";
	qDebug() << "QVCameraThread::run() <- return";
	}

/********************* Auxiliary QVMPlayerIOProcessor class *******************/
QVMPlayerIOProcessor::QVMPlayerIOProcessor(QProcess * mplayer): QObject(), command_queue(), command_queue_mutex()
	{
	qDebug() << "QVMPlayerIOProcessor::QVMPlayerIOProcessor()";
	rows = cols = 0;
	fps = time_length = time_pos = 0;
	speed = 1;
	this->mplayer = mplayer;

	queueCommandToMPlayer("get_time_length");

	qDebug() << "QVMPlayerIOProcessor::QVMPlayerIOProcessor() <- return";
	}

QVMPlayerIOProcessor::~QVMPlayerIOProcessor()
	{
	qDebug() << "QVMPlayerIOProcessor::~QVMPlayerIOProcessor()";
	qDebug() << "QVMPlayerIOProcessor::~QVMPlayerIOProcessor() <- return";
	}

void QVMPlayerIOProcessor::queueCommandToMPlayer(const QString &com_str, bool ifEmpty)
	{
	qDebug() << "QVMPlayerIOProcessor::queueCommand("<< com_str <<")";
	qDebug() << "QVMPlayerIOProcessor::queueCommand(): command_queue.size() = "
		<< command_queue.size();
	command_queue_mutex.lock();
	if(!ifEmpty || command_queue.size()==0)
		command_queue << com_str + '\n';
	command_queue_mutex.unlock();
	qDebug() << "QVMPlayerIOProcessor::queueCommand() <- return";
	}

void QVMPlayerIOProcessor::sendCommandToMPlayer()
	{
	qDebug() << "QVMPlayerIOProcessor::sendCommandToMPlayer()";
	command_queue_mutex.lock();
	if (command_queue.size() > 0)
		{
		qDebug() << "QVMPlayerIOProcessor::sendCommandToMPlayer(): sending command:" << command_queue.head();
		mplayer->write(qPrintable(command_queue.dequeue()));
		}
	command_queue_mutex.unlock();
	qDebug() << "QVMPlayerIOProcessor::sendCommandToMPlayer(): wait for bytes written";
	// BUG: In Windows this hangs?
	#ifdef unix
	mplayer->waitForBytesWritten();
	#endif
	qDebug() << "QVMPlayerIOProcessor::sendCommandToMPlayer() <- return";
	}

int QVMPlayerIOProcessor::interpretMPlayerOutput()
	{
	qDebug() << "QVMPlayerIOProcessor::interpretMPlayerOutput()";
	int length = -1;
	char buf[1024];

	length = mplayer->readLine(buf, sizeof(buf));

	if (length == -1)
		qDebug() << "QVMPlayerIOProcessor::interpretMPlayerOutput(): length == -1";
	else	{
		QString str(buf);
		QStringList variables = str.simplified().split("=");
		QStringList palabras = str.simplified().split(" ");

		if(variables[0] == "ANS_speed")
			{
			speed = variables[1].toDouble();
			qDebug() << "QVMPlayerIOProcessor::interpretMPlayerOutput(): updating speed =" << speed;
			}
		else if(variables[0] == "ANS_LENGTH")
			{
			time_length = variables[1].toDouble();
			qDebug() << "QVMPlayerIOProcessor::interpretMPlayerOutput(): updating time_length =" << time_length;
			}
		else if(variables[0] == "ANS_TIME_POSITION")
			{
			time_pos = variables[1].toDouble();
			qDebug() << "QVMPlayerIOProcessor::interpretMPlayerOutput(): updating time_pos =" << time_pos;
			}
		else if (palabras[0] == "VIDEO:")
			{
			// We have FPS specification
			for (int i = 0; i < palabras.size(); ++i)
				if (palabras[i] == "fps")
					fps = palabras[i-1].toDouble();

			qDebug() << "QVMPlayerIOProcessor::interpretMPlayerOutput(): updating fps" << fps;
			}
		else if (palabras[0] == "VO:")
			{
			// We have video specifications and image size.
			QStringList dimensiones = palabras[2].split("x");
			cols = dimensiones[0].toInt();
			rows = dimensiones[1].toInt();
			qDebug() << "QVMPlayerIOProcessor::interpretMPlayerOutput(): updating cols" << cols;
			qDebug() << "QVMPlayerIOProcessor::interpretMPlayerOutput(): updating rows" << rows;
			}
		else
			qDebug() << "QVMPlayerIOProcessor::interpretMPlayerOutput(): uninterpreted mplayer output:" << str;
		}
	qDebug() << "QVMPlayerIOProcessor::interpretMPlayerOutput() <- return" << length;
	return length;
	}

/******************* Auxiliary QVMPlayerFrameGrabber Class ********************/
QVMPlayerFrameGrabber::QVMPlayerFrameGrabber(QVMPlayerIOProcessor *mplayerIOProcessor, QString pipeName,
	bool realTime, const uInt waitMilisecs): QObject()
	{
	// We create a guarding thread that will unblock the subsequent fifo_file.readLine in the
	// case that mplayer does not start OK (due, for example, to a wrong URL, an incorrect
	// file format, or whatever). If mplayer starts OK before two seconds (for local videos) or
	// ten seconds (for DVD, VCD and remote videos), this thread silently stops and is deleted.
	QVCheckOKMplayerCamera check_thread(fifoInput, waitMilisecs);

	// We open the pipeline file with the designated name:
	fifoInput.setFileName(pipeName);
	if(fifoInput.open(QIODevice::ReadWrite|QIODevice::Unbuffered) == -1)
		qFatal("Error opening fifo %s\n", qPrintable(pipeName));

	// Now we read YUV header:
	int cols = 0, rows = 0, fps = 0;
	finished = !readYUV4MPEG2Header(fifoInput, cols, rows, fps);

	check_thread.quit();
	check_thread.wait();

	// Case the header is incorrect, stop.
	if (finished)
		return;

	// If the header was correctly read, proceed to init MPlayer IO processing.
	qDebug() << "QVMPlayerFrameGrabber::QVMPlayerFrameGrabber(fifoInput,"
			<< mplayerIOProcessor->rows << "x" << mplayerIOProcessor->cols << "," << realTime <<")";

	//std::cout << "xxx QVMPlayerFrameGrabber::QVMPlayerFrameGrabber(fifoInput,"
	//		<< mplayerIOProcessor->rows << "x" << mplayerIOProcessor->cols << "," << realTime <<")" << std::endl;
	
	this->mplayerIOProcessor = mplayerIOProcessor;
	this->realTime = realTime;
	this->frames_read = 0;

	//debug_in_memcpy = debug_in_updateSlot = false;

	/// @todo Es necesario inicializar esta imagen?
	imgY = QVImage<uChar>(cols, rows, cols);

	img_auxY = QVImage<uChar>(cols, rows, cols);
	img_auxU = QVImage<uChar>(cols/2, rows/2, cols/2);
	img_auxV = QVImage<uChar>(cols/2, rows/2, cols/2);

	if (realTime)
		{
		qvcamerathread = new QVCameraThread(this,SLOT(updateSlot()));
		qvcamerathread->start(QThread::NormalPriority);
		}

	connect(this, SIGNAL(sendCommandSignal()), mplayerIOProcessor, SLOT(sendCommandToMPlayer()));
	qDebug() << "QVMPlayerFrameGrabber::QVMPlayerFrameGrabber() <- return";
	};

QVMPlayerFrameGrabber::~QVMPlayerFrameGrabber()
	{
	qDebug() << "QVMPlayerFrameGrabber::~QVMPlayerFrameGrabber()";
	fifoInput.close();

	finished = true;
	if (realTime) delete qvcamerathread;
	qDebug() << "QVMPlayerFrameGrabber::~QVMPlayerFrameGrabber(): thread deleted.";
	qDebug() << "QVMPlayerFrameGrabber::~QVMPlayerFrameGrabber() <- return";
	}

void QVMPlayerFrameGrabber::updateSlot()
	{
	qDebug() << "QVMPlayerFrameGrabber::updateSlot()";

	qDebug() << "QVMPlayerFrameGrabber::updateSlot(): sent command to mplayer: get_time_pos";
	mplayerIOProcessor->queueCommandToMPlayer("pausing_keep get_time_pos\n",true);

	// Avoid QSocketNotifier annoying message, by deferring call:
	emit sendCommandSignal(); // NOT mplayerIOProcessor->sendCommandToMPlayer();

	// New frame read:
	qDebug() << "QVMPlayerFrameGrabber::updateSlot(): reading frame";
	QVImage<uChar, 1> tempY;
	QVImage<uChar, 1> tempU;
	QVImage<uChar, 1> tempV;

	// We always read in temporal images, that are later safely copied:
	// There is no copy overhead because QVImage uses QSharedData (reference
	// counting).
	tempY = img_auxY;
	tempU = img_auxU;
	tempV = img_auxV;

	qDebug() << "QVMPlayerFrameGrabber::readYUV4MPEG2Frame(): reading YUV header";
	finished = !readYUV4MPEG2Frame(fifoInput, tempY, tempU, tempV);

	// If in real time, we lock, to avoid collision with the thread that is
	// grabbing from the camera:
	if(realTime)
		mutex.lock();

	img_auxY = tempY;
	img_auxU = tempU;
	img_auxV = tempV;

	// Unlock (if in real time):
	if(realTime)
		{
		condition.wakeAll(); // Wake possible waiting grabs.
		mutex.unlock();
		}

	qDebug() << "QVMPlayerFrameGrabber::updateSlot(): finished" << finished;
	frames_read++;
	qDebug() << "QVMPlayerFrameGrabber::updateSlot(): frames read" << frames_read;

	mplayerIOProcessor->interpretMPlayerOutput();

	qDebug() << "QVMPlayerFrameGrabber::updateSlot() <- return";

	emit newReadFrameGrabber();
	}

void QVMPlayerFrameGrabber::updateFrameBuffer()
	{
	qDebug() << "QVMPlayerFrameGrabber::updateFrameBuffer()";

	// Here we come from a grab, so we always lock (if in real time, it is 
	// obviously needed, but also if not in real time, to make non-real-cameras
	// grabbing also thread-safe (won't be used normally, but who knows...
	// maybe the programmer wish to read different frames from different
	// threads, without discarding any frame as could happen with real time
	// cameras).
	mutex.lock();

	if(realTime)
		condition.wait(&mutex); // If in real time, wait for next fresh frame.

	// If in real time, updateSlot is asynchronously called by a dedicated
	// thread, otherwise we have to call it explicitly.
	if(not realTime)
		updateSlot();

	imgY = img_auxY;
	imgU = img_auxU;
	imgV = img_auxV;

	mutex.unlock();

	qDebug() << "QVMPlayerFrameGrabber::updateFrameBuffer() <- return";
	}

void QVMPlayerFrameGrabber::getQVImageYUV(QVImage<uChar> &imgY, QVImage<uChar> &imgU, QVImage<uChar> &imgV) const
	{
	qDebug() << "QVMPlayerFrameGrabber::getQVImageYUV()";
	imgY = this->imgY;
	imgU = this->imgU;
	imgV = this->imgV;
	qDebug() << "QVMPlayerFrameGrabber::getQVImageYUV() <- return";
	}

void QVMPlayerFrameGrabber::getQVImageGray(QVImage<uChar> &imgGray) const
	{
	qDebug() << "QVMPlayerFrameGrabber::getQVImageGray()";
	imgGray = imgY;
	qDebug() << "QVMPlayerFrameGrabber::getQVImageGray() <- return";
	}

/***************************** QVMPlayerProxy Class **************************/
QVMPlayerProxy::QVMPlayerProxy(QString /* name */): status(QVCamera::Closed), open_options(Default)
	{
	qDebug() << "QVMPlayerProxy::QVMPlayerProxy()";

	frames_grabbed = 0;
	live_camera = FALSE;

	qDebug() << "QVMPlayerProxy::QVMPlayerProxy() <- return";
	}

QVMPlayerProxy::~QVMPlayerProxy()
	{
	qDebug() << "QVMPlayerProxy::~QVMPlayerProxy()";
	if (status != QVCamera::Closed) closeCam();
	qDebug() << "QVMPlayerProxy::~QVMPlayerProxy() <- deleting image";
	qDebug() << "QVMPlayerProxy::~QVMPlayerProxy() <- return";
	}

void QVMPlayerProxy::initMPlayerArgs(QString urlString, uInt rows, uInt cols)
	{
	qDebug() << "QVMPlayerProxy::initMPlayerArgs()";
	qDebug() << "QVMPlayerProxy::initMPlayerArgs(): url string =" << urlString;

	// Loop option and fixed-vo must go first:
	if(not (open_options & NoLoop)) mplayer_args << "-loop" << "0";

	mplayer_args << "-fixed-vo";

	QUrl url(urlString);

	if (url.host() != "")
		path = url.host() + "/";

	path += url.path();

	qDebug() << "QVMPlayerProxy::initMPlayerArgs(): path" << path;

	// Here we infer the type of the video source from the url.
	// If it is not specified in the schema of it, certain match rules are
	// applied to guess it out.
	if (url.scheme() != "")				// schema specified by the user
		schema = url.scheme();
	else if (urlString.startsWith("/dev/video"))	// a v4l device
		schema = "v4l";
	else if (urlString.startsWith("/dev/dv"))	// a dv device
		schema = "dv";
	else if (urlString.contains("*"))		// a multifile
		schema = "mf";
	else if (urlString.startsWith("www."))		// a http file
		schema = "http";
	else if (urlString.startsWith("ftp."))		// a ftp file
		schema = "ftp";

	qDebug() << "QVMPlayerProxy::initMPlayerArgs(): scheme" << schema;

	live_camera = TRUE;	// default configuration

	// Different mplayer args for different kind of image sources:
	if ((schema == "v4l") or (schema == "v4l2") or (schema == "analog"))
		{
		// Video for Linux cameras:
		QString urlQueryValues = QString("driver=%1:device=%2").arg(schema).arg(path);

		QList<QPair<QString, QString> > queryItems = url.queryItems();
		for (int i = 0; i < queryItems.size(); ++i)
			urlQueryValues += ":" + queryItems.at(i).first + "=" + queryItems.at(i).second;

		mplayer_args << "tv://" << "-tv" << urlQueryValues;
		}
	else if (schema == "dv")
		// DV cameras (connected through firewire):
		mplayer_args << path << "-demuxer" << "rawdv" << "-cache" << "400";
	else if (schema == "iidc")
		// IIDC cameras (connected through firewire):
		// For example, iidc:///dev/video1394/0?from=/dev/video0:to=/dev/video1
		qFatal("Currently this driver does not work (apparently with\n"
			"vloopback writing and then reading from a fifo with mplayer\ndoes not work).\n");
	else if (schema == "tv")
		// Analog TV input:
		qFatal("tv URL: Still not implemented\n");
	else if (schema == "dvb")
		// Digital TV input:
		qFatal("dvb URL: Still not implemented\n");
	else	// Format can be rtsp:// http://, ftp://, dvd://, vcd://, mf:// or file://
		{ 
		//  We pass the url it directly to mplayer:
		live_camera = FALSE;
		if (schema != "")
			mplayer_args << QString(schema + "://" + path);
		else
			mplayer_args << path;
		}

	// IMPORTANT!! All -vf filters MUST be together, separated by commas, and in
	// the right order. By now, we only use deinterlacing and scale, and in that order:
	QString aux;

	// Deinterlace option:
	if(open_options & Deinterlaced) aux = "pp=md";

	if(rows != 0 or cols != 0)
		{
		if(aux != QString())
			aux += QString(",scale=%1:%2").arg(cols).arg(rows);
        	else
			aux = QString("scale=%1:%2").arg(cols).arg(rows);
        	}

	if (aux != QString()) mplayer_args << "-vf" << aux;

	// Real time option:
	if(not (open_options & RealTime))
		{
		if(not live_camera)
        		mplayer_args << "-benchmark";
		else
			// The user attempted to open a live camera in continuous mode.
			// We issue a warning:
			qWarning("Warning: opening live cameras in continuous mode "
					 "wastes less CPU time, but it is prone to delays in "
					 "the pipe of images if your process is slower than "
					 "the camera.");
		}

	// Additional arguments (slave, quiet, nosound & yuv4mpeg output):
	mplayer_args << "-slave" << "-quiet" << "-nosound" << "-vo" << QString("yuv4mpeg:file=%1").arg(namedPipe->getInputFilePath());

	qDebug() << "QVMPlayerProxy::initMPlayerArgs() <- return";
	}

// Main method to open a camera:
bool QVMPlayerProxy::openCam(const QString & urlstring, OpenOptions opts, uInt suggestedRows, uInt suggestedCols)
	{
	//std::cout << "xxx QVMPlayerProxy::openCam(" << qPrintable(urlstring) << ","
	//	<< suggestedCols << "x" << suggestedRows << "," << opts << ")" << std::endl;

	qDebug() << "QVMPlayerProxy::openCam(" << urlstring << ","
		<< suggestedCols << "x" << suggestedRows << "," << opts << ")";

	setStatus(QVCamera::Closed);

	open_options = opts;

	// Pipe initialization
	namedPipe = new QNamedPipe(QString("mplayer"));

	initMPlayerArgs(urlstring, suggestedRows, suggestedCols);

	//std::cout << "xxx QVMPlayerProxy::openCam(): MPlayer command: " << MPLAYER_BINARY_PATH;
	//foreach(QString arg, mplayer_args)
	//	std::cout << " " << qPrintable(arg);
	//std::cout << std::endl;

	qDebug() << "QVMPlayerProxy::openCam(): MPlayer args ->" << mplayer_args;
	mplayer.start(MPLAYER_BINARY_PATH, mplayer_args);
	qDebug() << "QVMPlayerProxy::openCam(): after mplayer.start()";
	if(not mplayer.waitForStarted(1000))
		qFatal("Mplayer failed to start: Are you sure it is installed and in the correct PATH?");
	qDebug() << "QVMPlayerProxy::openCam(): after mplayer.waitForstarted()";
	
	qDebug() << "QVMPlayerProxy::openCam(): creating QVMPlayerIOProcessor...";
	mplayerIOProcessor = new QVMPlayerIOProcessor(&mplayer);
	// -----------------------------------------

	const uInt waitMilisecs = (	schema == "http" or schema == "ftp" or
					schema == "rtsp" or schema == "dvd" or
					schema == "vcd" )? 10000:2000;

	// Important to open in ReadWrite mode, to avoid infinite waiting if mplayer did
	// not start OK, and because perhaps the former guarding thread will write in the FIFO.
	qDebug() << "QVMPlayerProxy::openCam(): opening fifo_file";

	mplayerFrameGrabber = new QVMPlayerFrameGrabber(mplayerIOProcessor, namedPipe->getOutputFilePath(),
							open_options & QVMPlayerProxy::RealTime, waitMilisecs);

	if (mplayerFrameGrabber->isFinished())
		{ 
		std::cerr << "QVMPlayerProxy::openCam(): Warning: Mplayer could not open the requested video source ("
			<< qPrintable(urlstring) << ") of type " << qPrintable(schema) <<std::endl;

		mplayer.terminate();
		if (not mplayer.waitForFinished(500)) mplayer.kill();	
		delete namedPipe;

		qDebug() << "QVMPlayerProxy::openCam(): QVCheckOKMplayerCamera thread deleted after detecting an error."
			<< " Returning FALSE";
		return FALSE;
		}

	qDebug() << "QVMPlayerProxy::openCam(): QVCheckOKMplayerCamera thread deleted after correct launch of mplayer";


	// -----------------------------------------
	qDebug() << "QVMPlayerProxy::openCam(): initializing MPlayer.";

	// These signals are to defer the call to sendCommandToMplayer, and avoid
	// the "QSocketNotifier:: socket notifier cannot be enabled from another
	// thread" warning.
	connect(mplayerFrameGrabber,SIGNAL(newReadFrameGrabber()),this,SIGNAL(newRead()));
	connect(this,SIGNAL(sendCommandSignal()),mplayerIOProcessor,SLOT(sendCommandToMPlayer()));

	mplayer.waitForReadyRead();

	setStatus(QVCamera::Running);

	qDebug() << "QVMPlayerProxy::openCam() <- return";

	emit camOpened();

	return TRUE;
	}

// Close camera method:
void QVMPlayerProxy::closeCam()
	{
	qDebug() << "QVMPlayerProxy::closeCam()";

	if (status == QVCamera::Closed)
		{
		qDebug() << "QVMPlayerProxy::closeCam(): camera already closed. Returning";
		qDebug() << "QVMPlayerProxy::closeCam() <- return";
		return;
		}

	setStatus(QVCamera::Closed);

	emit camClosed();

	mplayerIOProcessor->queueCommandToMPlayer("quit");	
	// Avoid QSocketNotifier annoying message, by deferring call:
	emit sendCommandSignal();

	delete mplayerFrameGrabber;
	qDebug() << "QVMPlayerProxy::closeCam(): mplayerFrameGrabber deleted";

	delete mplayerIOProcessor;
	qDebug() << "QVMPlayerProxy::closeCam(): mplayerIOProcessor deleted";

	mplayer.terminate();
	while (not mplayer.waitForFinished(500)) mplayer.kill();
	qDebug() << "QVMPlayerProxy::closeCam(): mplayer terminated";
	
	delete namedPipe;
	qDebug() << "QVMPlayerProxy::closecam(): deleted namedpipe";

	qDebug() << "QVMPlayerProxy::closeCam() <- return";
	}

bool QVMPlayerProxy::grab(QVImage<uChar> &imgY, QVImage<uChar> &imgU, QVImage<uChar> &imgV)
	{
	qDebug() << "QVMPlayerProxy::grab(imageY, imageU, imageV)";

	bool newImages = performGrab();

	if (newImages)
		mplayerFrameGrabber->getQVImageYUV(imgY, imgU, imgV);

	qDebug() << "QVMPlayerProxy::grab() <~ return";

	return newImages;
	}

bool QVMPlayerProxy::grab(QVImage<uChar,3> & imageRGB)
	{
	qDebug() << "QVMPlayerProxy::grab(imageRGB)";
	bool newImages = performGrab();

	QVImage<uChar> grabbedYImage, grabbedUImage, grabbedVImage;

	if (newImages)
		{
		mplayerFrameGrabber->getQVImageYUV(grabbedYImage, grabbedUImage, grabbedVImage);

		imageRGB = QVImage<uChar,3>(grabbedYImage.getCols(),grabbedYImage.getRows());
		YUV420ToRGB(grabbedYImage, grabbedUImage, grabbedVImage, imageRGB);
		}

	qDebug() << "QVMPlayerProxy::grab() <~ return";
	return newImages;
	}

bool QVMPlayerProxy::grab(QVImage<uChar,1> & imageGray)
	{
	qDebug() << "QVMPlayerProxy::grab(imageGray)";

	bool newImages = performGrab();

	if (newImages)
		mplayerFrameGrabber->getQVImageGray(imageGray);

	qDebug() << "QVMPlayerProxy::grab() <~ return";
	return newImages;
	}

bool QVMPlayerProxy::performGrab()
	{
	// First, if there are any pending commands, we send them to mplayer:
	qDebug() << "QVMPlayerProxy::performGrab()";

	if (isClosed()) return false;

	qDebug() << "QVMPlayerProxy::performGrab(): status != Closed";

	bool newFrameGrabbed = false;
	// This is for when the camera is paused; it will return the same frames
	// over and over, but not immediately, but at a maximum fps rate (to avoid
	// intense CPU usage, that will slow the rest of the application and the
	// GUI.

	switch(status)
		{
		case QVCamera::RunningOneStep:
			qDebug() << "QVMPlayerProxy::performGrab(): status == RunningOneStep";
			setStatus(QVCamera::Paused);
		case QVCamera::Running:
			qDebug() << "QVMPlayerProxy::performGrab(): status == Running";
			mplayerFrameGrabber->updateFrameBuffer();
			frames_grabbed++;
			newFrameGrabbed = true;
			qDebug() << "QVMPlayerProxy::performGrab(): frames grabbed" << getFramesGrabbed();
			break;

		case QVCamera::Paused:
			qDebug() << "QVMPlayerProxy::performGrab(): status == Paused";
			/*
			mutexJustToSleepMiliseconds.lock();
			qDebug() << "QVMPlayerProxy::performGrab(): pausing" << milisecsToWait
				<< "milisecs (video is "<< getFPS() <<" FPS's)";
			conditionJustToSleepMiliseconds.wait(&mutexJustToSleepMiliseconds,milisecsToWait);
			qDebug() << "QVMPlayerProxy::performGrab(): unlocking";
			mutexJustToSleepMiliseconds.unlock(); */
			break;

		default:
			break;
		}
	qDebug() << "QVMPlayerProxy::performGrab(): checking finished";

	if (mplayerFrameGrabber->isFinished())
		closeCam();
	else
		emit newGrab();

	qDebug() << "QVMPlayerProxy::performGrab() <- return";

	return newFrameGrabbed;
	}

// Pause, unpause, next frame, set speed, seek and close camer slots:
// In all of them, the corresponding command is sent to the shared command queue
// data structure, so access is accordingly protected .
void QVMPlayerProxy::pauseCamSlot()
	{
	qDebug() << "QVMPlayerProxy::pauseCam()";
	if(status == QVCamera::Closed) return;
	setStatus(QVCamera::Paused);
	qDebug() << "QVMPlayerProxy::pauseCam() <- return";
	}

void QVMPlayerProxy::unpauseCamSlot()
	{
	qDebug() << "QVMPlayerProxy::unpauseCam()";
	if(status == QVCamera::Closed) return;
	setStatus(QVCamera::Running);
	qDebug() << "QVMPlayerProxy::unpauseCam() <~ return";
	}

void QVMPlayerProxy::nextFrameCamSlot()
	{
	qDebug() << "QVMPlayerProxy::nextFrameCam()";
	if(status == QVCamera::Closed) return;
	setStatus(QVCamera::RunningOneStep);
	qDebug() << "QVMPlayerProxy::nextFrameCam() <~ return";
	}

void QVMPlayerProxy::setSpeedCamSlot(double d)
	{
	qDebug() << "QVMPlayerProxy::setSpeedCam()";
	if(status == QVCamera::Closed) return;
	mplayerIOProcessor->queueCommandToMPlayer(QString("pausing_keep speed_set ") + QString::number(d));
	mplayerIOProcessor->queueCommandToMPlayer("get_property speed");
	qDebug() << "QVMPlayerProxy::setSpeedCam() <~ return";
	}

void QVMPlayerProxy::seekCamSlot(QVCamera::TSeekType seek,double d)
	{
	qDebug() << "QVMPlayerProxy::seekCam()";
	if(status == QVCamera::Closed) return;
	if(status == QVCamera::Paused)
		setStatus(QVCamera::RunningOneStep);
	QString command = QString("pausing_keep seek ") + QString::number(d) + " " + QString::number(seek);
	mplayerIOProcessor->queueCommandToMPlayer(command);
	qDebug() << "QVMPlayerProxy::seekCam() <~ return";
	}
