/*
 *	Copyright (C) 2007, 2008, 2009. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvfunction.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVFUNCTION_H
#define QVFUNCTION_H

/*!
@class QVFunction qvfunction.h qvmath/qvfunction.h
@brief Base class for function objects.

This class can be used to create function class types. Objects derived from this class will wrap functions so they can be provided as parameters to other methods or procedures.

These objects can be used as any C++ function, using their overloaded function operator (). This operator calls to the virtual method @ref evaluate, which must be implemented by subclasses of @ref QVFunction to contain the code of the function:

@code
[...]
class FactorialFunction: public QVFunction<int, int>
	{
	public:
		int evaluate(const int &value)
			{
			int accum = 1;
			for (int i = 0; i < value; i++)
				accum *= i;
			return accum;
			}
	}
[...]
int main()
	{
	[...]
	FactorialFunction factorial;
	std::cout << "10! = " << factorial(10) << std::endl;
	[...]
	}
@endcode

Function objects can contain parametrization data to evaluate the function, besides each given input value:

@code
class QuadraticFunction: public QVFunction<double, double>
	{
	private:
		const double a, b, c;

	public:
		QuadraticFunction(const double a, const double b, const double c):
			QVFunction<double, double>(), a(a), b(b), c(c)
			{ }

		double evaluate(const double &x) const
			{ return a*x*x + b*x + c; }
	};
@endcode

@ingroup qvnumericalanalysis
*/
template <typename Input, typename Output> class QVFunction
	{
	public:
		QVFunction()	{ }
		virtual Output evaluate(const Input &)	const = 0;
		Output operator()(const Input &input)	const { return evaluate(input); }
	};
#endif
