/*
 *	Copyright (C) 2007, 2008, 2009. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

#include <qvprojective.h>
#include <qvmatrixalgebra.h>
#include <qvdefines.h>
#include <qvmath.h>
#include <float.h>
#include <qvnumericalanalysis.h>

/// @file qvprojective.cpp
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.
void homogenizePoints(const QList< QPair<QPointF, QPointF> > &matchings,
	QVMatrix &premult, QVMatrix &postmult,
	QList< QPair<QPointF, QPointF> > &homogeneizedPairs
	)
	{
	// 0. Fast normalization (similar to DLT, but based on min and max values
	// of X and Y, instead of mean value and normalized mean distance to
	// centroid equal to sqrt(2). Hartley-Zisserman 2nd ed. pag 109.
	float minXS=FLT_MAX, maxXS=-FLT_MAX, minYS=FLT_MAX, maxYS=-FLT_MAX,
	      minXD=FLT_MAX, maxXD=-FLT_MAX, minYD=FLT_MAX, maxYD=-FLT_MAX;

	QPair<QPointF, QPointF> matching;
	foreach(matching, matchings)
		{
		minXS = MIN(matching.first.x(), minXS);
		maxXS = MAX(matching.first.x(), maxXS);
		minYS = MIN(matching.first.y(), minYS);
		maxYS = MAX(matching.first.y(), maxYS);

		minXD = MIN(matching.second.x(), minXD);
		maxXD = MAX(matching.second.x(), maxXD);
		minYD = MIN(matching.second.y(), minYD);
		maxYD = MAX(matching.second.y(), maxYD);
		}

	// This is to avoid possible div by zero in degenerate conditions:
	if(fabs(minXS-maxXS) < EPSILON)
		maxXS += 1.0;
	if(fabs(minYS-maxYS) < EPSILON)
		maxYS += 1.0;
	if(fabs(minXD-maxXD) < EPSILON)
		maxXD += 1.0;
	if(fabs(minYD-maxYD) < EPSILON)
		maxYD += 1.0;

	minXS = 0; maxXS = 320;
	minYS = 0; maxYS = 240;

	//QPair<QPointF, QPointF> matching;
	foreach(matching, matchings)
		{
		const double	x = (matching.first.x()-(maxXS+minXS)/2)/(maxXS-minXS),
				y = (matching.first.y()-(maxYS+minYS)/2)/(maxYS-minYS),
				x_p = (matching.second.x()-(maxXD+minXD)/2)/(maxXD-minXD),
				y_p = (matching.second.y()-(maxYD+minYD)/2)/(maxYD-minYD);

		homogeneizedPairs.append(QPair<QPointF, QPointF>(QPointF(x,y),QPointF(x_p,y_p)));
		}

	premult = QVMatrix(3,3), postmult = QVMatrix(3,3);
	premult(0,0) = 1/(maxXS-minXS);	premult(0,1) = 0;		premult(0,2) = -(maxXS+minXS)/(2*(maxXS-minXS));
	premult(1,0) = 0;		premult(1,1) = 1/(maxYS-minYS);	premult(1,2) = -(maxYS+minYS)/(2*(maxYS-minYS));
	premult(2,0) = 0;		premult(2,1) = 0;		premult(2,2) = 1;

	postmult(0,0) = maxXD-minXD;	postmult(0,1) = 0;		postmult(0,2) = (maxXD+minXD)/2;
	postmult(1,0) = 0;		postmult(1,1) = maxYD-minYD;	postmult(1,2) = (maxYD+minYD)/2;
	postmult(2,0) = 0; 		postmult(2,1) = 0; 		postmult(2,2) = 1;
	}

QVMatrix ComputeProjectiveHomography(const QList< QPair<QPointF, QPointF> > &matchings)
	{
	Q_ASSERT(matchings.size() >= 4);

	// 0. Fast normalization (similar to DLT, but based on min and max values
	// of X and Y, instead of mean value and normalized mean distance to
	// centroid equal to sqrt(2). Hartley-Zisserman 2nd ed. pag 109.
	QList< QPair<QPointF, QPointF> > homogeneizedPairs;
	QVMatrix premult, postmult;

	homogenizePoints(matchings, premult, postmult, homogeneizedPairs);

	// 1. Obtain coeficient matrix for the system:
	//	Ax = 0
	QVMatrix coefsMatrix(3*homogeneizedPairs.size(),9);
	for (int n = 0; n < homogeneizedPairs.size(); n++)
		{
		const QPair<QPointF, QPointF> matching = homogeneizedPairs.at(n);
		const double	x = matching.first.x(),
				y = matching.first.y(),
				x_p = matching.second.x(),
				y_p = matching.second.y();

		coefsMatrix(3*n, 0) = 0;		coefsMatrix(3*n, 1) = 0; 		coefsMatrix(3*n, 2) = 0;
		coefsMatrix(3*n, 3) = -x;		coefsMatrix(3*n, 4) = -y; 		coefsMatrix(3*n, 5) = -1;
		coefsMatrix(3*n, 6) = x*y_p;		coefsMatrix(3*n, 7) = y*y_p; 		coefsMatrix(3*n, 8) = y_p;

		coefsMatrix(3*n+1, 0) = x;		coefsMatrix(3*n+1, 1) = y; 		coefsMatrix(3*n+1, 2) = 1;
		coefsMatrix(3*n+1, 3) = 0;		coefsMatrix(3*n+1, 4) = 0; 		coefsMatrix(3*n+1, 5) = 0;
		coefsMatrix(3*n+1, 6) = -x*x_p;		coefsMatrix(3*n+1, 7) = -y*x_p;		coefsMatrix(3*n+1, 8) = -x_p;

		coefsMatrix(3*n+2, 0) = -x*y_p;		coefsMatrix(3*n+2, 1) = -y*y_p;		coefsMatrix(3*n+2, 2) = -y_p;
		coefsMatrix(3*n+2, 3) = x*x_p;		coefsMatrix(3*n+2, 4) = y*x_p; 		coefsMatrix(3*n+2, 5) = x_p;
		coefsMatrix(3*n+2, 6) = 0;		coefsMatrix(3*n+2, 7) = 0; 		coefsMatrix(3*n+2, 8) = 0;
		}

	// 2. Solve the homogeneous system.
	QVVector x(9);

	solveHomogeneousLinear(coefsMatrix, x);

	// 3. Rearrange elements of vector 'x' in a 3x3 matrix.
	QVMatrix homography(3,3);
	homography(0,0) = x[0];	homography(0,1) = x[1];	homography(0,2) = x[2];
	homography(1,0) = x[3];	homography(1,1) = x[4];	homography(1,2) = x[5];
	homography(2,0) = x[6];	homography(2,1) = x[7];	homography(2,2) = x[8];

	// 4. De-normalize homography back:
	homography = (postmult * homography) * premult;
	homography = homography / homography(2,2);

	return homography;
	}

QVMatrix ComputeFundamentalMatrix(const QList< QPair<QPointF, QPointF> > &matchings)
	{
	Q_ASSERT(matchings.size() >= 4);

	///////////////////////////////////////////////////////
	// 0. Fast normalization (similar to DLT, but based on min and max values
	// of X and Y, instead of mean value and normalized mean distance to
	// centroid equal to sqrt(2). Hartley-Zisserman 2nd ed. pag 109.
	QList< QPair<QPointF, QPointF> > homogeneizedPairs;
	QVMatrix premult, postmult;

	homogenizePoints(matchings, premult, postmult, homogeneizedPairs);

	// 1. Obtain coeficient matrix for the system:
	//	Ax = 0
	QVMatrix coefsMatrix(homogeneizedPairs.size(),9);
	for (int n=0; n<homogeneizedPairs.size(); n++)
		{
		const QPair<QPointF, QPointF> matching = homogeneizedPairs.at(n);
		const double	x = matching.first.x(),
				y = matching.first.y(),
				x_p = matching.second.x(),
				y_p = matching.second.y();

		coefsMatrix(n, 0) = x*x_p;	coefsMatrix(n, 1) = y*x_p; 	coefsMatrix(n, 2) = x_p;
		coefsMatrix(n, 3) = x*y_p;	coefsMatrix(n, 4) = y*y_p; 	coefsMatrix(n, 5) = y_p;
		coefsMatrix(n, 6) = x;		coefsMatrix(n, 7) = y;	 	coefsMatrix(n, 8) = 1;
		}

	// 2. Solve the homogeneous system.
	QVVector x(9);
	solveHomogeneousLinear(coefsMatrix, x);

	// 3. Rearrange elements of vector 'x' in a 3x3 matrix.
	QVMatrix homography(3,3);
	homography(0,0) = x[0];	homography(0,1) = x[1];	homography(0,2) = x[2];
	homography(1,0) = x[3];	homography(1,1) = x[4];	homography(1,2) = x[5];
	homography(2,0) = x[6];	homography(2,1) = x[7];	homography(2,2) = x[8];

	// 4. De-normalize homography back:
	homography = (postmult * homography) * premult;
	homography = homography / homography(2,2);

	return homography;
	}

void getMeanDirection(const QVMatrix m, QVVector &mean, QVVector &direction)
	{
	mean = m.meanCol();
	QVMatrix centeredM = m;
	for (int i = 0; i < centeredM.getCols(); i++)
		centeredM.setCol(i, centeredM.getCol(i) - mean);

	// 1. Compute main direction vector.
	QVMatrix eigVecs;
	QVVector eigVals;
	eigenDecomposition(centeredM * centeredM.transpose(), eigVals, eigVecs);

	direction = QVVector(eigVecs.getCols());
	for (int i = 0; i < eigVecs.getCols(); i++)
		direction[i] = eigVecs(0,i);
	direction = direction * eigVals[0];
	}

// Aux functions
/*template <typename T1, typename T2> QList<T1> getFirstPairList(const QList< QPair<T1,T2> > &list)
	{
	QList<T1> result;
	QPair<T1,T2> pair;
	foreach(pair, list)
		result.append(pair.first);
	return result;
	}

/// @ingroup qvdta
/// @todo document this
template <typename T1, typename T2> QList<T2> getSecondPairList(const QList< QPair<T1,T2> > &list)
	{
	QList<T2> result;
	QPair<T1,T2> pair;
	foreach(pair, list)
		result.append(pair.second);
	return result;
	}*/

QVMatrix ComputeEuclideanHomography(const QList< QPair<QPointF, QPointF> > &matchings)
	{
	// 1. Get source and destination mean point and direction.
	QList<QPointF> sourcePointList, destPointList;
	QPair<QPointF, QPointF> matching;
	foreach(matching, matchings)
		{
		sourcePointList.append(matching.first);
		destPointList.append(matching.second);
		}

	const QVMatrix sources = sourcePointList, destinations = destPointList;

	QVVector sourcesMean, destinationsMean, sourcesDirection, destinationsDirection;
	getMeanDirection(sources.transpose(), sourcesMean, sourcesDirection);
	getMeanDirection(destinations.transpose(), destinationsMean, destinationsDirection);

	const QPointF C1 = sourcesMean, C2 = destinationsMean, D1 = sourcesDirection, D2 = destinationsDirection;

	// 2. Get zoom and angle
	double zoom = 0;
	int switchD1Direction = 0, zoomCount = 0;
	for(int i = 0; i < sourcePointList.size(); i ++)
		{
		const QPointF sourceCenteredPoint = sourcePointList.at(i) - sourcesMean, destCenteredPoint = destPointList.at(i) - destinationsMean;

		if (norm2(sourceCenteredPoint) > 1e-10)
			{
			zoom += norm2(destCenteredPoint) / norm2(sourceCenteredPoint);
			zoomCount ++;
			}

		if (	(norm2(destCenteredPoint - D2) - norm2(destCenteredPoint + D2)) * (norm2(sourceCenteredPoint + D1)
				- norm2(sourceCenteredPoint - D1)) > 0	|| 
			(norm2(destCenteredPoint - D2) - norm2(destCenteredPoint + D2)) * (norm2(sourceCenteredPoint + D1)
				- norm2(sourceCenteredPoint - D1)) > 0	)
			switchD1Direction++;
		else
			switchD1Direction--;
		}

	zoom /= sourcePointList.size();

	const double delta = qvClockWiseAngle((switchD1Direction > 0)?-D1:D1, D2);

	// 3. Get the euclidean homography matrix, derived from the following code for Mathematica:
	//	Rotation[delta_] := { {Cos[delta], Sin[delta], 0}, {-Sin[delta], Cos[delta], 0}, {0, 0, 1} };
	//	Translation[x_, y_] := { {1, 0, x}, {0, 1, y}, {0, 0, 1} };
	//	Zoom[z_] := { {z, 0, 0}, {0, z, 0}, {0, 0, 1} };
	//	result = Translation[C2_x, C2_y].Zoom[zoom].Rotation[delta].Translation[C1_x, C1_y] // MatrixForm
	QVMatrix result = QVMatrix::identity(3);
	result(0,0) = zoom*cos(delta);	result(0,1) = zoom*sin(delta);	result(0,2) = C2.x() - zoom*cos(delta)*C1.x() - zoom*C1.y()*sin(delta);
	result(1,0) = -zoom*sin(delta);	result(1,1) = zoom*cos(delta);	result(1,2) = C2.y() - zoom*cos(delta)*C1.y() + zoom*C1.x()*sin(delta);

	return result;
	}

QPointF ApplyHomography(const QVMatrix &H, const QPointF &point)
	{
	const double	h11 = H(0,0), h12 = H(0,1), h13 = H(0,2),
			h21 = H(1,0), h22 = H(1,1), h23 = H(1,2),
			h31 = H(2,0), h32 = H(2,1), h33 = H(2,2),
			x = point.x(), y = point.y(),
			homogenizer = h31*x + h32*y + h33;

	return QPointF(h11*x + h12*y + h13, h21*x + h22*y + h23)/homogenizer;
	}

QList<QPointF> ApplyHomography(const QVMatrix &homography, const QList<QPointF> &sourcePoints)
	{
	QList<QPointF> result;
	foreach(QPointF point, sourcePoints)
		result.append(ApplyHomography(homography, point));
	return result;
	}

double HomographyTestError(const QVMatrix &homography)
	{
	const QVVector v1 = homography.getCol(0), v2 = homography.getCol(1);
	return	ABS(v1.norm2() - v2.norm2()) / (v1.norm2() + v2.norm2())
		+ ABS(v1 * v2) / (v1.norm2() * v2.norm2());
	}

////////////////////////////////////////////////////////////////
// Focal calibration from homography
//

// To get direct focal distance formula:
//
//	K = {{f, 0, 0}, {0, f, 0}, {0, 0, 1}};
//	Hvar = {{h1, h2, h3}, {h4, h5, h6}, {h7, h8, h9}};
//	ErrorMat[H_, K_] := Transpose[H].Inverse[K].Inverse[K].H;
//	Error[ErrorM_] := ((ErrorM[[1, 1]] - ErrorM[[2, 2]])^2 + ErrorM[[1, 2]]^2) / ErrorM[[1, 1]]^2;
//	ErrorFunction[H_, K_] := Error[ErrorMat[H, K]];
//	P = Dt[ErrorFunction[Hvar, K], f, Constants -> {h1, h2, h3, h4, h5, h6, h7, h8, h9}] ;
//	focalDistance  = Solve[P == 0, f][[2]][[1]] // FullSimplify
//	CForm[focalDistance]
//
void GetDirectIntrinsicCameraMatrixFromHomography(const QVMatrix &H, QVMatrix &K)
	{
	const double h1 = H(0,0), h2 = H(0,1), h4 = H(1,0), h5 = H(1,1), h7 = H(2,0), h8 = H(2,1);
	const double focalNumerator =	+ (h2*(h2 + h4) - h1*h5 + pow(h5,2))*(pow(h2,2) - h2*h4 + h5*(h1 + h5))*pow(h7,2)
					- (pow(h1,2) + pow(h4,2))*(h1*h2 + h4*h5)*h7*h8
					+ (pow(h1,2) + pow(h4,2))*(pow(h1,2) - pow(h2,2) + pow(h4,2) - pow(h5,2))*pow(h8,2);
	const double focalDenominator = + (pow(h2,2) + pow(h5,2))*pow(h7,4)
					- (h1*h2 + h4*h5)*pow(h7,3)*h8
					- (pow(h2,2) + pow(h5,2))*pow(h7,2)*pow(h8,2)
					+ (pow(h1,2) + pow(h4,2))*pow(h8,4);
	const double finv = sqrt(ABS(focalNumerator / focalDenominator))/2;

	K = QVMatrix::identity(3) * finv;
	K(2,2) = 1;
	}

void CalibrateCameraFromPlanarHomography(const QVMatrix &H, QVMatrix &K, QVMatrix &Rt)
	{
	// 3.2. Compute intrinsic camera matrix
	// K is such that:
	// H = K*Rt
	GetDirectIntrinsicCameraMatrixFromHomography(H, K);

	// ----------------------------------------------------
	// 3.3. Compute extrinsic camera matrix
	// 3.3.1. Eliminate K matrix from H
	QVMatrix R12t = pseudoInverse(K)*H;

	// 3.3.2. Tricky: homogenize M3x3 such as the lenght of rotation vectors is the unit.
	// Divide M3x3 by the mean of the square values for first two elements
	// of the diagonal of matrix M3x3^t * M3x3
	Rt = QVMatrix(4,4);
	R12t = R12t * 2 / (R12t.getCol(0).norm2() + R12t.getCol(1).norm2());

	// 3.3.3. Compose matrix R and -R*C vector in final matrix M
	Rt.setCol(0, R12t.getCol(0));
	Rt.setCol(1, R12t.getCol(1));
	Rt.setCol(2, R12t.getCol(0) ^ R12t.getCol(1));
	Rt.setCol(3, R12t.getCol(2));
	Rt(3,3) = 1;
	}

//
// This function decomposes the matrix for an homography, obtaining matrix R and vector T from the decomposition:
//	H = K*M3x3
// where
//	M3x3 = [ R1 | R2 | -R^t*C ]
// obtaining the essencial matrix M of size 3x4:
//		[	|	 ]
//	M4x4 =	[  Rot  | -R^t*C ]
//		[ _____	| ______ ]
//		[ 0 0 0	|    1   ]	
//
void GetExtrinsicCameraMatrixFromHomography(const QVMatrix &K, const QVMatrix &H, QVMatrix &Rt)
	{
	// 1. Eliminate K matrix from H
	QVMatrix R12_t = pseudoInverse(K)*pseudoInverse(H);

	// 2. Tricky: homogenize M3x3 such as the lenght of rotation vectors is the unit.
	//	2.2 Divide M3x3 by the mean of the square values for first two elements
	//	of the diagonal of matrix M3x3^t * M3x3
	R12_t = R12_t * 2 / (R12_t.getCol(0).norm2() + R12_t.getCol(1).norm2());

	// 3. Compose matrix R and -R*C vector in final matrix M
	Rt = QVMatrix(4,4);

	Rt.setCol(0, R12_t.getCol(0));
	Rt.setCol(1, R12_t.getCol(1));
	Rt.setCol(2, R12_t.getCol(0) ^ R12_t.getCol(1));
	Rt.setCol(3, R12_t.getCol(2));
	Rt(3,3) = 1;
	}

void GetPinholeCameraIntrinsicsFromPlanarHomography(const QVMatrix &H, QVMatrix &K, const int iterations,
						const double maxGradientNorm, const double step, const double tol)
	{
	class KErrorFunction: public QVFunction<QVVector, double>
		{
		private:
			const QVMatrix H;
	
			double evaluate(const QVVector &input) const
				{
				const QVMatrix	Rt = pseudoInverse(KMatrix(input))*H, errorMat = Rt.transpose() * Rt;
				return POW2(errorMat(0,0) -1) + POW2(errorMat(1,1) -1) + 2*POW2(errorMat(1,0));
				}
	
		public:
			KErrorFunction(const QVMatrix &H): QVFunction<QVVector, double>(), H(H)		{ }
	
			const QVMatrix KMatrix(const QVVector &input) const
				{
				QVMatrix K = QVMatrix::zeros(3,3);
				K(0,0) = input[0];
				K(1,1) = input[0];
				K(2,2) = input[1];
				return K;
				}
		} errorFunction(H);

	QVVector x(2,1);
	qvGSLMinimizeFDF(errorFunction, x, VectorBFGS, iterations, maxGradientNorm, step, tol);
	K = errorFunction.KMatrix(x);
	K = K / K(2,2);
	}
